PROGRAM DNAPars (infile, treefile, OUTPUT);
   (* version 3.31. (c) Copyright 1986, 1990 by the University of Washington 
      and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
      granted to copy and use this program provided no fee is charged for it 
      and provided that this copyright notice is not removed. *)

CONST maxsp = 25;    (* maximum number of species               *)
   maxsz = 49;       (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 1500;  (* max. number base sites per sequence     *)
   maxpatterns = 512;  (* max number of different site patterns *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of tied trees stored     *)
   maxuser = 8;      (* maximum number of user-defined trees *)

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
   stepshortarray = ARRAY[1..maxpatterns] OF INTEGER;
   sequence = ARRAY[1..maxsp] OF ARRAY [1..maxsites] OF CHAR;
   bases = (A, C, G, U, O);
   baset = SET OF bases;
   basearray = ARRAY [1..maxpatterns] OF baset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      base : basearray; (* the sequence *)
      numsteps : stepshortarray; (* bookkeeps steps *)
      visited : BOOLEAN; (* for use in counting steps *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;
   longer = ARRAY[0..5] OF INTEGER;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, endsite, inseed, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        outgrno indicates outgroup *)
   jumble, usertree, weights, thresh, trout, outgropt, anerror : BOOLEAN;
   weight, oldweight, alias, ally, location : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   y : sequence;
   threshold : REAL;
   threshwt : stepshortarray;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;
   basechar : ARRAY[1..31] OF CHAR;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
      (* test whether ch is a digit *)
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
              OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
              OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a') 
                      OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE charbase;
      (* make ARRAY basechar to hold characters for base sets using a
         binary coding with O as most significant position, A least *)
      VAR i : INTEGER;
   BEGIN (* charbase *)
      basechar[1] := 'A';
      basechar[2] := 'C';
      basechar[3] := 'M';
      basechar[4] := 'G';
      basechar[5] := 'R';
      basechar[6] := 'S';
      basechar[7] := 'V';
      basechar[8] := 'T';
      basechar[9] := 'W';
      basechar[10] := 'Y';
      basechar[11] := 'H';
      basechar[12] := 'K';
      basechar[13] := 'D';
      basechar[14] := 'B';
      basechar[15] := 'N';
      basechar[16] := 'O';
      FOR i := 17 TO 31 DO
         basechar[i] := '?';
   END; (* charbase *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'DNA parsimony algorithm, version 3.31');
         WRITELN(OUTPUT);
         READ(infile, spp, chars);
         WRITELN(OUTPUT, spp:2,' species, ',chars:3,'  sites');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxsites
         THEN WRITELN(OUTPUT, 'TOO MANY SITES: adjust CONSTants');
         anerror := (spp > maxsp) OR (chars > maxsites);
         WRITELN(OUTPUT);
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputoutgroup;
            (* input the number of the outgroup *)
         BEGIN (* inputoutgroup *)
            READLN(infile, outgrno);
            IF (outgrno < 1) OR (outgrno > spp)
            THEN BEGIN
               WRITELN(OUTPUT, 'BAD OUTGROUP NUMBER: ', outgrno:4);
               anerror := TRUE;
               END;
         END; (* inputoutgroup *)

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            thresh := TRUE;
            READLN(infile, threshold);
            IF threshold < 1.0
            THEN BEGIN
               WRITELN(OUTPUT, 'BAD THRESHOLD VALUE');
               anerror := TRUE;
               END
            ELSE threshold := TRUNC(threshold*10.0 + 0.5)/10.0;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '   Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE BEGIN
                     WRITELN(OUTPUT, 'BAD WEIGHT CHARACTER: ',ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'   Sites are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         jumble := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         thresh := FALSE;
         trout := FALSE;
         usertree := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'J') OR (ch = 'O') OR (ch = 'T')
               OR (ch = 'U') OR (ch = 'W') OR (ch = 'Y')
            THEN CASE ch OF
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'T' : extranum := extranum + 1;
                  'U' : usertree := TRUE;
                  'W' : extranum := extranum + 1;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'J'
               THEN READLN(infile, inseed);
               IF ch = 'O'
               THEN inputoutgroup;
               IF ch = 'T'
               THEN inputhresh;
               IF ch = 'W'
               THEN inputweights;
               anerror := NOT ((ch = 'J') OR (ch = 'O') OR (ch = 'T')
                               OR (ch = 'W'));
               IF anerror
               THEN WRITELN(OUTPUT, 'ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               END;
            END;
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT, 'Random number seed = ', inseed:8);
            WRITELN(OUTPUT);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread : BOOLEAN;
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + ((chars + ((chars-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         WRITE(OUTPUT, 'Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Sequences');
         WRITE(OUTPUT, '----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '---------');
         WRITELN(OUTPUT);
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            FOR i := 1 TO spp DO
               IF NOT anerror
               THEN BEGIN
                  IF basesread = 0
                  THEN FOR j := 1 TO nmlngth DO
                        IF NOT anerror
                        THEN BEGIN
                           READ(infile, name[i,j]);
                           anerror := anerror OR EOF(infile) OR EOLN(infile);
                           IF anerror
                           THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                               'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                           END;
                  j := basesread;
                  WHILE (j < chars) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                        THEN BEGIN
                           WRITELN(OUTPUT, 'ERROR: BAD BASE:', charstate,
                              ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        j := j + 1;
                        IF charstate = '.' THEN charstate := y[1][j];
                        y[i][j] := charstate;
                        END;
                     END;
                  IF i = 1 THEN basesnew := j;
                  READLN(infile);
                  anerror := anerror OR (j <> basesnew);
                  IF anerror
                  THEN WRITELN(OUTPUT, 'ERROR: SEQUENCES OUT OF ALIGNMENT');
                  END;
            basesread := basesnew;
            allread := basesread = chars;
            END;
         IF NOT anerror
         THEN BEGIN
            FOR i := 1 TO ((chars-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO spp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(OUTPUT, name[j,k]);
                  WRITE(OUTPUT, '   ');
                  l := 60*i;
                  IF l > chars THEN l := chars;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j][k] = y[1][k])
                     THEN charstate := '.'
                     ELSE charstate := y[j][k];
                     WRITE(OUTPUT, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort keeping sites, weights in same order *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := chars DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO chars-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     tied := TRUE;
                     k := 1;
                     WHILE (k <= spp) AND tied DO BEGIN
                        flip := y[k][jj] > y[k][jg];
                        tied := tied AND (y[k][jj] = y[k][jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        itemp := weight[j];
                        weight[j] := weight[j+gap];
                        weight[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < chars DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= chars) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= spp) AND tied DO BEGIN
                     tied := tied AND (y[k][alias[i]] = y[k][alias[j]]);
                     k := k + 1;
                     END;
                  IF tied
                  THEN BEGIN
                     weight[i] := weight[i] + weight[j];
                     weight[j] := 0;
                     ally[alias[j]] := alias[i];
                     END;
                  j := j + 1;
                  END;
              i := j - 1;
              END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so one representative of each pattern of
               sites comes first *)
            VAR i, j, itemp : INTEGER;
               done, found : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               IF NOT (ally[alias[i]] = alias[i])
               THEN BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= chars
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := (ally[alias[j]] = alias[j]);
                        j := j + 1;
                     UNTIL found OR (j > chars);
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        itemp := weight[i];
                        weight[i] := weight[j];
                        weight[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               i := i + 1;
               done := done OR (i >= chars);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO chars DO BEGIN
            alias[i] := i;
            oldweight[i] := weight[i];
            ally[i] := i;
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         endsite := 0;
         FOR i := 1 TO chars DO
            IF ally[i] = i
            THEN endsite := endsite + 1;
         FOR i := 1 TO endsite DO
            location[alias[i]] := i;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN(OUTPUT, 'TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
         FOR i := 1 TO endsite DO BEGIN
            weight[i] := 10*weight[i];
            threshwt[i] := TRUNC(threshold * weight[i] + 0.5);
            END;
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j : INTEGER;
            ns : baset;
      BEGIN (* makevalues *)
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            treenode[i]^.tip := (i <= spp);
            treenode[i]^.visited := FALSE;
            END;
         FOR j := 1 TO endsite DO BEGIN
            FOR i := 1 TO spp DO BEGIN
               CASE y[i][alias[j]] OF
                  'A' : ns := [A];
                  'C' : ns := [C];
                  'G' : ns := [G];
                  'U' : ns := [U];
                  'T' : ns := [U];
                  'M' : ns := [A, C];
                  'R' : ns := [A, G];
                  'W' : ns := [A, U];
                  'S' : ns := [C, G];
                  'Y' : ns := [C, U];
                  'K' : ns := [G, U];
                  'B' : ns := [C, G, U];
                  'D' : ns := [A, G, U];
                  'H' : ns := [A, C, U];
                  'V' : ns := [A, C, G];
                  'N' : ns := [A, C, G, U];
                  'X' : ns := [A, C, G, U];
                  '?' : ns := [A, C, G, U];
                  'O' : ns := [O];
                  '-' : ns := [O]
                  END;
               treenode[i]^.base[j] := ns;
               treenode[i]^.numsteps[j] := 0;
               END;
            END;
      END; (* makevalues *)

   BEGIN (* doinput *)
      inputnumbers;
      IF NOT anerror
      THEN inputoptions;
      IF NOT anerror
      THEN inputdata;
      IF NOT anerror
      THEN makeweights;
      IF NOT anerror
      THEN makevalues;
   END; (* doinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      TYPE placearray = ARRAY[1..maxsz] OF INTEGER;
      VAR i, j, k, numtrees, num, nextree, which, minwhich, minsteps : INTEGER;
         like, bestyet, bestlike, gotlike, wt, sumw, sum, sum2, sd : REAL;
         lastrearr : BOOLEAN;
         nsteps : ARRAY [1..maxtrees] OF REAL;
         fsteps : ARRAY [1..maxuser] OF ARRAY [1..maxpatterns] OF REAL;
         there, item, nufork, dummy : nodeptr;
         place : placearray;
         bestrees : ARRAY[1..maxtrees] OF placearray;

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
         VAR p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
         p := newfork;
         IF root = below THEN root := newfork;
         root^.ancestor := NIL;
         REPEAT
            p^.visited := FALSE;
            bottom := (p = root);
            p := p^.ancestor;
         UNTIL bottom;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below, p : nodeptr;
            bottom : BOOLEAN;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               END;
            fork^.ancestor := NIL;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            IF root = fork THEN root := other;
            IF root <> other
            THEN BEGIN
               p := below;
               REPEAT
                  p^.visited := FALSE;
                  bottom := (p = root);
                  p := p^.ancestor;
               UNTIL bottom;
               END;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, steps, sum, term : INTEGER;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (p: nodeptr);
               (* sets up for each node in the tree the base sequence
                  at that point and counts the changes.  The program
                  spends much of its time in this PROCEDURE *)
               VAR i : INTEGER;
                  ns, rs, ls : baset;
            BEGIN (* fillin *)
               WITH p^ DO
                  FOR i := 1 TO endsite DO BEGIN
                     ls := leftdesc^.base[i];
                     rs := rtdesc^.base[i];
                     ns := ls * rs;
                     numsteps[i] := leftdesc^.numsteps[i]
                        + rtdesc^.numsteps[i];
                     IF ns = []
                     THEN BEGIN
                        ns := ls + rs;
                        numsteps[i] := numsteps[i] + weight[i];
                        END;
                     base[i] := ns;
                     END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT (p^.tip OR p^.visited)
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               p^.visited := TRUE;
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         postorder (r);
         FOR i := 1 TO endsite DO BEGIN
            steps := root^.numsteps[i];
            IF steps <= threshwt[i]
            THEN term := steps
            ELSE term := threshwt[i];
            sum := sum + term;
            IF usertree AND (which <= maxtrees)
            THEN fsteps[which][i] := term;
            END;
         IF usertree
         THEN BEGIN
            nsteps[which] := sum;
            IF which = 1
            THEN BEGIN
               minwhich := 1;
               minsteps := sum;
               END
            ELSE IF sum < minsteps
               THEN BEGIN
                  minwhich := which;
                  minsteps := sum;
                  END;
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)
         TYPE knowed = RECORD    (* temporary storage for tree shape *)
            ancestor, leftdesc, rtdesc : nodeptr;
            END;
         VAR treeknowed : ARRAY [1..maxsz] OF knowed;

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR i, pos : INTEGER;
               found : BOOLEAN;
               rute : nodeptr;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               reroot (treenode[outgrno]);
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                  END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  WHILE (place[i] = bestrees[pos][i]) AND (i <= spp) DO
                     i := i + 1;
                  found := (i > spp);
                  below := place[i] < bestrees[pos][i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO
                  bestrees[i+1] := bestrees[i];
               bestrees[pos] := place;
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  FOR i := 1 TO nonodes DO BEGIN
                     treeknowed[i].ancestor := treenode[i]^.ancestor;
                     treeknowed[i].leftdesc := treenode[i]^.leftdesc;;
                     treeknowed[i].rtdesc := treenode[i]^.rtdesc;
                     END;
                  rute := root;
                  savetree;
                  FOR i := 1 TO nonodes DO BEGIN
                     treenode[i]^.ancestor := treeknowed[i].ancestor;
                     treenode[i]^.leftdesc := treeknowed[i].leftdesc;;
                     treenode[i]^.rtdesc := treeknowed[i].rtdesc;
                     END;
                  root := rute;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : nodeptr);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..nmlngth] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               END
            ELSE BEGIN
               FOR i := 1 TO nmlngth DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ',') OR (ch = ')') OR (n > nmlngth));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO nmlngth DO
                     found := found AND (str[i] = name[n, i]);
                  IF found
                  THEN p := treenode[n]
                  ELSE n := n + 1;
               UNTIL (n > spp) OR found;
               IF n > spp
               THEN BEGIN
                  WRITE(OUTPUT, 'Cannot find species: ');
                  FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[spp + 1];
         nextnode := spp;
         root^.ancestor := NIL;
         addelement (root);
         FOR i := 1 TO nonodes DO treenode[i]^.visited := FALSE;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '  ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF rover < over
                     THEN BEGIN
                        IF (i = (tipy-down)) AND (p^.xcoord = 0)
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF p^.index - spp >= 10
                        THEN WRITE(OUTPUT, (p^.index - spp):2)
                        ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '  remember:');
         IF outgropt
         THEN WRITE(OUTPUT, '(although rooted by outgroup)');
         WRITELN(OUTPUT, ' this is an unrooted tree!');
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each site *)
         VAR i, j, k, l : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, n : INTEGER;
               nothing : basearray;
               bottom : BOOLEAN;

            PROCEDURE ancestset (a, b : baset; VAR c : baset);
               (* make the set of ancestral states below nodes
                  whose base sets are a and b *)
            BEGIN (* ancestset *)
               c := a * b;
               IF c = []
               THEN c := a + b;
            END; (* ancestset *)

            PROCEDURE hyptrav (r : nodeptr; VAR hypset : basearray;
               b1, b2 : INTEGER);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  maybe, nonzero : BOOLEAN;
                  tempset, left, rt, anc : baset;
                  temparray, ancset : basearray;

               PROCEDURE hyprint (b1, b2 : INTEGER);
                  (* print out states in sites b1 through b2 at node *)
                  VAR i, j, k, n : INTEGER;
                     dot : BOOLEAN;
                     b : bases;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF NOT outgropt
                     THEN WRITE(OUTPUT, '       ')
                     ELSE WRITE(OUTPUT, 'root   ');
                     END
                  ELSE WRITE(OUTPUT, (r^.ancestor^.index - spp):4, '   ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, name[r^.index, i])
                  ELSE WRITE(OUTPUT, (r^.index - spp):4, '      ');
                  IF bottom
                  THEN WRITE(OUTPUT, '          ')
                  ELSE IF nonzero
                     THEN WRITE(OUTPUT, '   yes    ')
                     ELSE IF maybe
                        THEN WRITE(OUTPUT, '  maybe   ')
                        ELSE WRITE(OUTPUT, '   no     ');
                  FOR i := b1 TO b2 DO BEGIN
                     j := location[ally[i]];
                     tempset := r^.base[j];
                     anc := hypset[j];
                     IF NOT bottom
                     THEN anc := r^.ancestor^.base[j];
                     dot := (tempset = anc) AND (NOT bottom);
                     IF dot
                     THEN WRITE(OUTPUT, '.')
                     ELSE IF tempset = [A]
                        THEN WRITE(OUTPUT, 'A')
                        ELSE IF tempset = [C]
                           THEN WRITE(OUTPUT, 'C')
                           ELSE IF tempset = [G]
                              THEN WRITE(OUTPUT, 'G')
                              ELSE IF tempset = [U]
                                 THEN WRITE(OUTPUT, 'T')
                                 ELSE IF tempset = [O]
                                    THEN WRITE(OUTPUT, '-')
                                    ELSE BEGIN
                                       k := 1;
                                       n := 0;
                                       FOR b := A TO O DO BEGIN
                                          IF b IN tempset
                                          THEN n := n + k;
                                          k := k + k;
                                          END;
                                       WRITE(OUTPUT, basechar[n]);
                                       END;
                     IF (i MOD 10) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               maybe := FALSE;
               nonzero := FALSE;
               FOR i := b1 TO b2 DO BEGIN
                  j := location[ally[i]];
                  anc := hypset[j];
                  IF NOT r^.tip
                  THEN BEGIN
                     left := r^.leftdesc^.base[j];
                     rt := r^.rtdesc^.base[j];
                     tempset := left * rt * anc;
                     IF tempset = []
                     THEN BEGIN
                        tempset := (left * rt) + (left * anc) + (rt * anc);
                        IF tempset = []
                        THEN tempset := left + rt + anc;
                        END;
                     r^.base[j] := tempset;
                     END;
                  IF NOT bottom
                  THEN anc := r^.ancestor^.base[j];
                  nonzero := nonzero OR ((r^.base[j] * anc) = []);
                  maybe := maybe OR (r^.base[j] <> anc);
                  END;
               hyprint (b1, b2);
               bottom := FALSE;
               IF NOT r^.tip
               THEN BEGIN
                  temparray := r^.leftdesc^.base;
                  FOR i := b1 TO b2 DO BEGIN
                     j := location[ally[i]];
                     ancestset(hypset[j], r^.rtdesc^.base[j], ancset[j]);
                     END;
                  hyptrav (r^.leftdesc, ancset, b1, b2);
                  FOR i := b1 TO b2 DO BEGIN
                     j := location[ally[i]];
                     ancestset(hypset[j], temparray[j], ancset[j]);
                     END;
                  hyptrav (r^.rtdesc, ancset, b1, b2);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'From    To     Any Steps?    State at upper node');
            WRITE(OUTPUT, '                            ');
            WRITELN(OUTPUT, ' ( . means same as in the node below it on tree)');
            WRITELN(OUTPUT);
            FOR i := 1 TO endsite DO nothing[i] := [];
            FOR i := 1 TO ((chars-1) DIV 40)+1 DO BEGIN
               WRITELN(OUTPUT);
               bottom := TRUE;
               n := 40*i;
               IF n > chars THEN n := chars;
               hyptrav (root, nothing, 40*i-39, n);
               END;
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF name[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := name[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               END;
            IF p = root
            THEN WRITELN(treefile, ';');
         END; (* treeout *)

      BEGIN (* describe *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'requires a total of ',-like/10.0:10:3);
         WRITELN(OUTPUT);
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each site:');
         WRITE(OUTPUT, '      ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '     *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:5);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE BEGIN
                  l := location[ally[k]];
                  IF oldweight[k] > 0
                  THEN WRITE(OUTPUT,
                     (oldweight[k]*(root^.numsteps[l] DIV weight[l])):4)
                  ELSE WRITE(OUTPUT, '   0');
                  END;
               END;
            WRITELN(OUTPUT);
            END;
         hypstates;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         IF trout
         THEN treeout (root);
      END; (* describe *)

   BEGIN (* maketree *)
      charbase;
      IF NOT usertree
      THEN BEGIN
         FOR i := 1 TO spp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO spp DO BEGIN
            j := TRUNC(randum(seed)*spp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         root := treenode[enterorder[1]];
         add (treenode[enterorder[1]],
            treenode[enterorder[2]], treenode[spp+1]);
         WRITELN(OUTPUT, 'Adding species:');
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, name[enterorder[1],i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, name[enterorder[2],i]);
         WRITELN(OUTPUT);
         lastrearr := FALSE;
         FOR i := 3 TO spp DO BEGIN
            bestyet := -32000.0;
            item := treenode[enterorder[i]];
            nufork := treenode[spp + i - 1];
            addpreorder (root, item, nufork);
            add (there, item, nufork);
            like := bestyet;
            rearrange (root);
            WRITE(OUTPUT, '   ');
            FOR j := 1 TO nmlngth DO
               WRITE(OUTPUT, name[enterorder[i],j]);
            WRITELN(OUTPUT);
            lastrearr := (i = spp);
            IF lastrearr
            THEN BEGIN
               WRITELN(OUTPUT, 'Doing global rearrangements');
               WRITE(OUTPUT, '  !');
               FOR j := 1 TO nonodes DO WRITE(OUTPUT, '-');
               WRITELN(OUTPUT, '!');
               bestlike := bestyet;
               nextree := 1;
               REPEAT
                  WRITE(OUTPUT, '   ');
                  gotlike := bestlike;
                  FOR j := 1 TO nonodes DO BEGIN
                     there := root;
                     bestyet := -32000.0;
                     item := treenode[j];
                     IF item <> root
                     THEN BEGIN
                        nufork := treenode[j]^.ancestor;
                        remove (item, nufork);
                        there := root;
                        addpreorder (root, item, nufork);
                        add (there, item, nufork);
                        END;
                     WRITE(OUTPUT, '.');
                     END;
                  WRITELN(OUTPUT);
               UNTIL bestlike <= gotlike;
               END;
            END;
         FOR i := spp DOWNTO 2 DO
            remove (treenode[i], dummy);
         WRITELN(OUTPUT);
         IF (nextree = 2)
         THEN WRITELN(OUTPUT, 'One most parsimonious tree found:')
         ELSE WRITELN(OUTPUT, nextree-1:6, ' trees in all found');
         IF nextree > maxtrees + 1
         THEN BEGIN
            WRITELN(OUTPUT, 'here are the first', maxtrees:4, ' of them');
            nextree := maxtrees + 1;
            END;
         WRITELN(OUTPUT);
         FOR i := 1 TO nextree-1 DO BEGIN
            root := treenode[1];
            add (treenode[1], treenode[2], treenode[spp+1]);
            FOR j := 3 TO spp DO
               add (treenode[bestrees[i][j]], treenode[j], treenode[spp+j-1]);
            reroot (treenode[outgrno]);
            evaluate (root);
            printree;
            describe;
            FOR j := 2 TO spp DO
               remove (treenode[j], dummy);
            END;
         IF trout
         THEN BEGIN
            WRITELN(OUTPUT, 'Trees also written onto file');
            WRITELN(OUTPUT);
            END;
         END
      ELSE BEGIN
         WRITE(OUTPUT, 'User-defined tree');
         READLN(infile, numtrees);
         IF numtrees > 1
         THEN WRITE(OUTPUT, 's');
         WRITELN(OUTPUT, ':');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            IF outgropt
            THEN reroot (treenode[outgrno]);
            evaluate (root);
            printree;
            describe;
            END;
         WRITELN(OUTPUT);
         IF (numtrees > 1) AND (chars > 1)
         THEN BEGIN
            WRITE(OUTPUT, 'Tree    Steps   Diff Steps   Its S.D.');
            WRITELN(OUTPUT, '   Significantly worse?');
            WRITELN(OUTPUT);
            IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
            FOR which := 1 TO num DO BEGIN
               WRITE(OUTPUT, which:3, nsteps[which]/10:10:1);
               IF minwhich = which
               THEN WRITELN(OUTPUT, '  <------ best')
               ELSE BEGIN
                  sumw := 0.0;
                  sum := 0.0;
                  sum2 := 0.0;
                  FOR j := 1 TO endsite DO
                     IF weight[j] > 0
                     THEN BEGIN
                        wt := weight[j]/10;
                        sumw := sumw + wt;
                        sum := sum + (fsteps[which][j]
                                      -fsteps[minwhich][j]) / 10;
                        sum2 := sum2 + SQR((fsteps[which][j]
                                            -fsteps[minwhich][j]) / 10) / wt;
                        END;
                  sd := SQRT((sumw/(sumw-1.0))*(sum2-SQR(sum/sumw)));
                  WRITE(OUTPUT, (nsteps[which]-minsteps)/10:10:1, sd:12:4);
                  IF sum > 1.95996*sd
                  THEN WRITELN(OUTPUT, '           Yes')
                  ELSE WRITELN(OUTPUT, '            No')
                  END;
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            END;
         END;
   END; (* maketree *)

BEGIN (* DNA parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   anerror := FALSE;
   doinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
      REWRITE(treefile);
      END;
   IF NOT anerror
   THEN maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);   Turbo Pascal and maybe others too *)
      END;
END. (* DNA parsimony by uphill search *)

