PROGRAM RestML (infile, treefile, OUTPUT);
  (* version 3.31. (c) Copyright 1986, 1990 by the University of Washington 
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
     granted to copy and use this program provided no fee is charged for it 
     and provided that this copyright notice is not removed. *)

   CONST maxsp = 10;     (* maximum number of species *)
      maxsp2 = 18;       (* 2 x maximum number of species - 2 *)
      maxsites = 250;    (* maximum number of sites *)
      maxpatterns = 100; (* maximum number of presence/absence patterns *)
      maxcutter = 8;     (* maximum number of bases in a site *)
      maxtrees = 8;      (* maximum number of user trees *)
      smoothings = 4;    (* number of passes in smoothing algorithm *)
      iterations = 4;    (* number of iterates of EM for each branch *)
      nmlngth = 10;      (* number of characters max. in species name *)
      epsilon = 0.00001; (* used in update *)
      extrapol = 1000.0; (* extrapolation factor to speed iteration *)
      initialv = 0.1;    (* starting value of branch length *)

   TYPE sitelike = ARRAY[0..maxcutter] OF REAL;
      phenotype = ARRAY [0..maxpatterns] OF sitelike;
      sequence = ARRAY[1..maxsp] OF ARRAY [1..maxsites] OF CHAR;
      naym = ARRAY[1..nmlngth] OF CHAR;
      longer = ARRAY[0..5] OF INTEGER;
      transmatrix = ARRAY[0..maxcutter, 0..maxcutter] OF REAL;
      transarray = ARRAY[1..maxsp2] OF transmatrix;
      nodeptr = ^node;
      node = RECORD
         next, back : nodeptr;
         tip        : BOOLEAN;
         branchnum,
            number  : INTEGER;
         x          : phenotype;
         nayme      : naym;
         v          : REAL;
         xcoord, ycoord, ymin, ymax : INTEGER
         END;
      tree = RECORD
         nodep      : ARRAY[1..maxsp2] OF nodeptr;
         trans,
          transprod : transarray;
         likelihood : REAL;
         start      : nodeptr
         END;

   VAR infile, treefile : TEXT;
      numsp, numsp1, numsp2, sites, endsite, weightsum,
         sitelength, inseed, outgrno : INTEGER;
      anerror, global, jumble, outgropt,
         weights, trout, trunc8, usertree : BOOLEAN;
      curtree, bestree : tree;
      y : sequence;
      seed : longer;
      enterorder : ARRAY[1..maxsp] OF INTEGER;
      weight, alias, aliasweight : ARRAY[0..maxsites] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   PROCEDURE getinput;
      (* read input file *)

      PROCEDURE getnums;
         (* read and print out numbers of species and sites *)
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,'Restriction site Maximum Likelihood');
         WRITELN(OUTPUT, ' method, version 3.31');
         WRITELN(OUTPUT);
         READ(infile, numsp, sites);
         WRITELN(OUTPUT, numsp:4, ' Species, ', sites:4, ' Sites');
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: adjust CONSTants');
         IF sites > maxsites
         THEN WRITELN(OUTPUT, 'TOO MANY SITES: adjust CONSTants');
         anerror := (numsp > maxsp) OR (sites > maxsites);
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      FUNCTION letter (ch : CHAR) : BOOLEAN;
         (* detects whether a character is a letter or not
            -- works for both ASCII and EBCDIC codes *)
      BEGIN (* letter *)
         letter := ((ch >= 'A') AND (ch <= 'I'))
                   OR ((ch >= 'J') AND (ch <= 'R'))
                   OR ((ch >= 'S') AND (ch <= 'Z'))
                   OR ((ch >= 'a') AND (ch <= 'i'))
                   OR ((ch >= 'j') AND (ch <= 'r'))
                   OR ((ch >= 's') AND (ch <= 'z'));
      END; (* letter *)

      PROCEDURE uppercase (VAR ch : CHAR);
         (* convert ch to upper case -- either ASCII or EBCDIC *)
      BEGIN (* uppercase *)
         IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                         OR  ('a' < 'A') AND (ch < 'A'))
         THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
      END; (* uppercase *)

      PROCEDURE getoptions;
         (* read the options information *)
         VAR ch : CHAR;
            i, extranum : INTEGER;

         PROCEDURE inputweights;
            (* input the character weights, 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            weightsum := 0;
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF (ch = '0') OR (ch = '1')
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE BEGIN
                  WRITELN(OUTPUT, 'BAD WEIGHT CHARACTER: ', ch,
                     ' -- WEIGHTS IN RESTML MUST BE 0 OR 1');
                  anerror := TRUE;
                  END;
               weightsum := weightsum + weight[i];
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'   Sites are weighted as follows:');
            WRITELN(OUTPUT);
            FOR i := 1 TO sites DO BEGIN
               IF ((i-1) MOD 60) = 0
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR j := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
                  END;
               WRITE(OUTPUT, weight[i]:1);
               IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
               THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* getoptions *)
         global := FALSE;
         jumble := FALSE;
         outgropt := FALSE;
         outgrno := 1;
         sitelength := 6;
         trout := FALSE;
         trunc8 := TRUE;
         usertree := FALSE;
         weights := FALSE;
         FOR i := 1 TO sites DO weight[i] := 1;
         weightsum := sites;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'G') OR (ch = 'J') OR (ch = 'L')
               OR (ch = 'O') OR (ch = 'U') OR (ch = 'W') OR (ch = 'Y')
            THEN CASE ch OF
                  'A' : trunc8 := FALSE;
                  'G' : global := TRUE;
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'L' : extranum := extranum + 1;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'U' : usertree := TRUE;
                  'W' : BEGIN
                        weights := TRUE;
                        extranum := extranum + 1;
                        END;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               anerror := NOT ((ch = 'J') OR (ch = 'O')
                            OR (ch = 'L') OR (ch = 'W'));
               IF anerror
               THEN WRITELN(OUTPUT, 'ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               READLN(infile);
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'J'
               THEN READ(infile, inseed);
               IF ch = 'O'
               THEN READ(infile, outgrno);
               IF (outgrno > numsp) OR (outgrno < 1)
               THEN WRITELN(OUTPUT, 'BAD OUTGROUP NUMBER: ', outgrno:6);
               anerror := anerror OR (outgrno > numsp) OR (outgrno < 1);
               IF ch = 'L'
               THEN READ(infile, sitelength);
               IF (sitelength > maxcutter) OR (sitelength < 4)
               THEN WRITELN(OUTPUT, 'BAD RESTRICTION SITE LENGTH: ',
                     sitelength:4);
               anerror := anerror OR (sitelength > maxcutter)
                  OR (sitelength < 4);
               IF ch = 'W'
               THEN inputweights;
               END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '  Recognition sequences all',
                           sitelength:2, ' bases long');
         IF trunc8
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'Sites absent from all species ',
                            'are assumed to have been omitted');
            WRITELN(OUTPUT);
            END;
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'Random number seed = ', inseed:8);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
         IF weights
         THEN printweights;
      END; (* getoptions *)

      PROCEDURE setuptree (VAR a : tree);
          (* set up data structures for a tree *)
         VAR i, j : INTEGER;
            p, q : nodeptr;
      BEGIN (* setuptree *)
         FOR i := 1 TO numsp DO BEGIN
            NEW(a.nodep[i]);
            a.nodep[i]^.tip := TRUE;
            a.nodep[i]^.number := i;
            END;
         FOR i := numsp1 TO numsp2 DO BEGIN
            q := NIL;
            FOR j := 1 TO 3 DO BEGIN
               NEW(p);
               p^.tip := FALSE;
               p^.number := i;
               p^.next := q;
               q := p;
               END;
            p^.next^.next^.next := p;
            a.nodep[i] := p;
            END;
         a.likelihood := -999999.0;
         a.start := a.nodep[1];
      END; (* setuptree *)

      PROCEDURE getdata;
         (* read the species and sites data *)
         VAR i, j, k, l, sitesread, sitesnew : INTEGER;
            ch : CHAR;
            allread : BOOLEAN;
      BEGIN (* getdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         WRITE(OUTPUT, 'Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Sites');
         WRITE(OUTPUT, '----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '-----');
         WRITELN(OUTPUT);
         setuptree (curtree);
         sitesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            FOR i := 1 TO numsp DO
               IF NOT anerror
               THEN BEGIN
                  IF sitesread = 0
                  THEN FOR j := 1 TO nmlngth DO
                        IF NOT anerror
                        THEN BEGIN
                           READ(infile, curtree.nodep[i]^.nayme[j]);
                           anerror := anerror OR EOF(infile) OR EOLN(infile);
                           IF anerror
                           THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                               'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                           END;
                  j := sitesread;
                  WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, ch);
                        uppercase (ch);
                        IF NOT ((ch = '1') OR (ch = '0') OR (ch = '+')
                           OR (ch = '-') OR (ch = '?') OR (ch = '.'))
                        THEN BEGIN
                           WRITELN(OUTPUT, ' WARNING -- BAD SYMBOL ', ch,
                                ' AT POSITION ', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        IF ch = '1'
                        THEN ch := '+';
                        IF ch = '0'
                        THEN ch := '-';
                        j := j + 1;
                        IF ch = '.' THEN ch := y[1][j];
                        y[i][j] := ch;
                        END;
                  IF i = 1 THEN sitesnew := j;
                  READLN(infile);
                  anerror := anerror OR (j <> sitesnew);
                  IF anerror
                  THEN WRITELN(OUTPUT, 'ERROR: SEQUENCES OUT OF ALIGNMENT');
                  END;
            sitesread := sitesnew;
            allread := sitesread = sites;
            END;
         IF NOT anerror
         THEN BEGIN
            FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO numsp DO BEGIN
                  FOR k := 1 TO nmlngth DO
                     WRITE(OUTPUT, curtree.nodep[j]^.nayme[k]);
                  WRITE(OUTPUT, '   ');
                  l := 60*i;
                  IF l > sites THEN l := sites;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     WRITE(OUTPUT, y[j][k]);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         IF (NOT usertree)
         THEN setuptree(bestree);
      END; (* getdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort keeping alias, aliasweight in same order *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := sites DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO sites-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     flip := FALSE;
                     tied := TRUE;
                     k := 1;
                     WHILE (k <= numsp) AND tied DO BEGIN
                        flip := y[k][jj] > y[k][jg];
                        tied := tied AND (y[k][jj] = y[k][jg]);
                        k := k + 1;
                        END;
                     IF tied
                     THEN BEGIN
                        aliasweight[j] := aliasweight[j] + aliasweight[j+gap];
                        aliasweight[j+gap] := 0;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        itemp := aliasweight[j];
                        aliasweight[j] := aliasweight[j+gap];
                        aliasweight[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < sites DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= sites) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= numsp) AND tied DO BEGIN
                     tied := tied AND (y[k][alias[i]] = y[k][alias[j]]);
                     k := k + 1;
                     END;
                  IF tied AND (aliasweight[j] > 0)
                  THEN BEGIN
                     aliasweight[i] := aliasweight[i] + aliasweight[j];
                     aliasweight[j] := 0;
                     alias[j] := alias[i];
                     END;
                  j := j + 1;
                  END;
              i := j - 1;
              END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so positively weighted sites come first *)
            VAR i, j, itemp : INTEGER;
               done, found : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               found := FALSE;
               IF aliasweight[i] > 0
               THEN i := i + 1
               ELSE BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= sites
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := aliasweight[j] > 0;
                        j := j + 1;
                     UNTIL found OR (j > sites);
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        itemp := aliasweight[i];
                        aliasweight[i] := aliasweight[j];
                        aliasweight[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               done := done OR (i >= sites);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO sites DO BEGIN
            alias[i] := i;
            aliasweight[i] := weight[i];
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         FOR i := 1 TO sites DO BEGIN
            weight[i] := aliasweight[i];
            IF weight[i] > 0
            THEN endsite := i;
            END;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN(OUTPUT, 'TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
         weight[0] := 1;
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j, k, l, m : INTEGER;
      BEGIN (* makevalues *)
         FOR k := 1 TO endsite DO BEGIN
            j := alias[k];
            FOR i := 1 TO numsp DO BEGIN
               FOR l := 0 TO sitelength DO
                  curtree.nodep[i]^.x[k][l] := 1.0;
               CASE y[i][j] OF
                  '+' : FOR m := 1 TO sitelength DO
                           curtree.nodep[i]^.x[k][m] := 0.0;
                  '-' : curtree.nodep[i]^.x[k][0] := 0.0;
                  '?' : BEGIN END
                  END;
               END;
            END;
         FOR i := 1 TO numsp DO BEGIN
            FOR k := 1 TO sitelength DO
               curtree.nodep[i]^.x[0][k] := 1.0;
            curtree.nodep[i]^.x[0][0] := 0.0;
            END;
      END; (* makevalues *)

   BEGIN (* getinput *)
      getnums;
      IF NOT anerror 
      THEN getoptions;
      IF NOT anerror
      THEN getdata;
      IF NOT anerror
      THEN makeweights;
      IF NOT anerror
      THEN makevalues;
   END; (* getinput *)

   PROCEDURE maketree;
      (* construct and rearrange tree *)
      VAR i, j, k, nextsp, numtrees, num, which, maxwhich : INTEGER;
         maxlogl, sum, sum2, sd : REAL;
         succeeded, smoothed : BOOLEAN;
         tempmatrix, tempprod : transmatrix;
         tempm, tempp : transarray;
         pi : sitelike;
         logl : ARRAY [1..maxtrees] OF REAL;
         logf : ARRAY [1..maxtrees] OF ARRAY [1..maxpatterns] OF REAL;

      PROCEDURE setuppi;
         (* set up equilibrium probabilities of being a given
            number of bases away from a restriction site *)
         VAR i : INTEGER;
            sum : REAL;
      BEGIN (* setuppi *)
         pi[0] := 1;
         sum := pi[0];
         FOR i := 1 TO sitelength DO BEGIN
            pi[i] := 3 * pi[i-1] * (sitelength - i + 1) / i;
            sum := sum + pi[i];
            END;
         FOR i := 0 TO sitelength DO
            pi[i] := pi[i] / sum;
      END; (* setuppi *)

      PROCEDURE maketrans (p : REAL);
         (* make transition matrix, product matrix with change
            probability p.  Put the results in tempmatrix, tempprod *)
         VAR i, j, k, m1, m2 : INTEGER;
            sump, sumn, pover3, pijk, nijk, f : REAL;
            binom1, binom2 : ARRAY[0..maxcutter] OF REAL;
      BEGIN (* maketrans *)
         pover3 := p / 3;
         f := 2*pover3/(1-pover3);
         FOR i := 0 TO sitelength DO BEGIN
            IF p > (1.0 - epsilon)
            THEN p := 1.0 - epsilon;
            binom1[0] := EXP((sitelength-i)*LN(1-p));
            FOR k := 1 TO sitelength-i DO
               binom1[k] := binom1[k-1] * (p/(1-p))
                               * (sitelength-i-k+1)/k;
            binom2[0] := EXP(i*LN(1-pover3));
            FOR k := 1 TO i DO
               binom2[k] := binom2[k-1] * (pover3/(1-pover3))
                             * (i-k+1)/k;
            FOR j := 0 TO sitelength DO BEGIN
               sump := 0.0;
               sumn := 0.0;
               IF (i-j) > 0
               THEN m1 := i-j
               ELSE m1 := 0;
               IF (sitelength - j) < i
               THEN m2 := sitelength - j
               ELSE m2 := i;
               FOR k := m1 TO m2 DO BEGIN
                  nijk := j-i + 2*k + f*(i-k);
                  pijk := binom1[j-i+k] * binom2[k];
                  sump := sump + pijk;
                  sumn := sumn + pijk*nijk;
                  END;
               tempmatrix[i,j] := sump;
               tempprod[i,j] := sumn;
               END;
            END;
      END; (* maketrans *)

      PROCEDURE branchtrans (i : INTEGER; p : REAL);
         (* make branch transition matrix, product matrices for branch
            i with probability of change p*)
      BEGIN (* branchtrans *)
         maketrans (p);
         curtree.trans[i] := tempmatrix;
         curtree.transprod[i] := tempprod;
      END; (* branchtrans *)

      FUNCTION evaluate (VAR tr : tree) : REAL;
         (* evaluates the likelihood, using info. at one branch *)
         VAR sum, sum2, y, liketerm, like0, lnlike0, term : REAL;
            i, j, k : INTEGER;
            p, q : nodeptr;
            x1, x2 : sitelike;
      BEGIN (* evaluate *)
         sum := 0.0;
         p := tr.start;
         q := p^.back;
         y := p^.v;
         maketrans (y);
         x1 := p^.x[0];
         x2 := q^.x[0];
         IF trunc8
         THEN BEGIN
            like0 := 0.0;
            FOR j := 0 TO sitelength DO BEGIN
               liketerm := pi[j] * x1[j];
               FOR k := 0 TO sitelength DO
                  like0 := like0 + liketerm * tempmatrix[j,k] * x2[k];
               END;
            lnlike0 := LN(1.0-like0);
            END;
         FOR i := 1 TO endsite DO BEGIN
            x1 := p^.x[i];
            x2 := q^.x[i];
            sum2 := 0.0;
            FOR j := 0 TO sitelength DO BEGIN
               liketerm := pi[j] * x1[j];
               FOR k := 0 TO sitelength DO
                  sum2 := sum2 + liketerm * tempmatrix[j,k] * x2[k];
               END;
            term := LN(sum2);
            IF trunc8
            THEN term := term - lnlike0;
            IF usertree AND (which <= maxtrees)
            THEN logf[which][i] := term;
            sum := sum + weight[i]*term;
            END;
         IF usertree
         THEN BEGIN
            logl[which] := sum;
            IF which = 1
            THEN BEGIN
               maxwhich := 1;
               maxlogl := sum;
               END
            ELSE IF sum > maxlogl
               THEN BEGIN
                  maxwhich := which;
                  maxlogl := sum;
                  END;
            END;
         tr.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE nuview (p : nodeptr);
         (* recompute fractional likelihoods for one part of tree *)
         VAR i, j, k, lowlim : INTEGER;
            sumq, sumr : REAL;
            q, r : nodeptr;
            xq, xr, xp : sitelike;
            tempq, tempr : transmatrix;
      BEGIN (* nuview *)
         IF trunc8
         THEN lowlim := 0
         ELSE lowlim := 1;
         q := p^.next^.back;
         r := p^.next^.next^.back;
         tempq := curtree.trans[q^.branchnum];
         tempr := curtree.trans[r^.branchnum];
         FOR i := lowlim TO endsite DO BEGIN
            xq := q^.x[i];
            xr := r^.x[i];
            FOR j := 0 TO sitelength DO BEGIN
               sumq := 0.0;
               sumr := 0.0;
               FOR k := 0 TO sitelength DO BEGIN
                  sumq := sumq + tempq[j,k] * xq[k];
                  sumr := sumr + tempr[j,k] * xr[k];
                  END;
               xp[j] := sumq * sumr;
               END;
            p^.x[i] := xp;
            END;
      END; (* nuview *)

      PROCEDURE update (p : nodeptr);
         (* improve branch length and views for one branch *)

         PROCEDURE makenewv (p : nodeptr);
            (* EM algorithm improvement of a branch length *)
            VAR i, j, k, lowlim, it : INTEGER;
               sum, sumlike, sumprod, liketerm, liket, y, yold, like0,
                  prod0, like, oldlike, olderlike, extrap : REAL;
               done : BOOLEAN;
               q : nodeptr;
               xx1, xx2 : sitelike;
         BEGIN (* makenewv *)
            extrap := extrapol;
            q := p^.back;
            y := p^.v;
            IF trunc8
            THEN lowlim := 0
            ELSE lowlim := 1;
            done := FALSE;
            oldlike := 0.0;
            olderlike := 0.0;
            it := 1;
            WHILE (it < iterations) DO BEGIN
               like := 0.0;
               maketrans(y);
               yold := y;
               sumlike := 0.0;
               FOR i := lowlim TO endsite DO BEGIN
                  xx1 := p^.x[i];
                  xx2 := q^.x[i];
                  sum := 0.0;
                  sumprod := 0.0;
                  FOR j := 0 TO sitelength DO BEGIN
                     liket := xx1[j]*pi[j];
                     FOR k := 0 TO sitelength DO BEGIN
                        liketerm := liket*xx2[k];
                        sumprod := sumprod + tempprod[j,k]*liketerm;
                        sum := sum + tempmatrix[j,k]*liketerm;
                        END;
                     END;
                  IF i = 0
                  THEN BEGIN
                     like0 := sum;
                     prod0 := sumprod;
                     END
                  ELSE BEGIN
                     sumlike := sumlike + weight[i] * sumprod / sum;
                     like := like + weight[i] * LN(sum);
                     END;
                  END;
               IF trunc8 AND (ABS(like0 - 1.0) > 1.0E-10)
               THEN like := like - weightsum * LN(1.0 - like0);
               y := sumlike / (sitelength*weightsum);
               IF (it >= 3) AND ((like-oldlike) < (oldlike-olderlike))
               THEN extrap := (oldlike - olderlike)
                               / (2.0*oldlike - olderlike - like) * extrapol
               ELSE extrap := extrapol;
               IF extrap < 0.0
               THEN extrap := extrapol;
               olderlike := oldlike;
               oldlike := like;
               IF trunc8
               THEN y := (1.0 - like0) * y + prod0 / sitelength;
               y := extrap*(y-yold)+yold;
               IF y < epsilon
               THEN y := 10.0*epsilon;
               IF y >= 0.75
               THEN y := 0.75;
               it := it + 1;
               END;
            smoothed := smoothed AND done;
            p^.v := y;
            q^.v := y;
            branchtrans (p^.branchnum, y);
         END; (* makenewv *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN nuview(p);
         IF NOT p^.back^.tip
         THEN nuview(p^.back);
         makenewv(p);
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
         (* update nodes throughout the tree, recursively *)
      BEGIN (* smooth *)
         update (p);
         IF NOT p^.tip
         THEN BEGIN
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE hookup (p, q : nodeptr);
         (* connect two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      PROCEDURE insert (p, q: nodeptr);
         (* insert a subtree into a branch, improve lengths in tree *)
         VAR i, m, n : INTEGER;
            r : nodeptr;
      BEGIN (* insert *)
         r := p^.next^.next;
         hookup (r, q^.back);
         hookup (p^.next, q);
         IF q^.v >= 0.75
         THEN q^.v := 0.75
         ELSE q^.v := -0.75 * (1 - EXP(-0.5*LN(1-1.333333*q^.v)));
         q^.back^.v := q^.v;
         r^.v := q^.v;
         r^.back^.v := r^.v;
         IF q^.branchnum = q^.number
         THEN BEGIN
            m := q^.branchnum;
            n := r^.number;
            END
         ELSE BEGIN
            m := r^.number;
            n := q^.branchnum;
            END;
         q^.branchnum := m;
         q^.back^.branchnum := m;
         r^.branchnum := n;
         r^.back^.branchnum := n;
         branchtrans (q^.branchnum, q^.v);
         branchtrans (r^.branchnum, r^.v);
         smoothed := FALSE;
         i := 1;
         WHILE (i < smoothings) AND (NOT smoothed) DO BEGIN
            smoothed := TRUE;
            smooth (p);
            smooth (p^.back);
            i := i + 1;
            END;
      END; (* insert *)

      PROCEDURE remove (VAR p, q : nodeptr);
         (* remove p and record in q where it was *)
      BEGIN (* remove *)
         q := p^.next^.back;
         hookup (q, p^.next^.next^.back);
         q^.back^.branchnum := q^.branchnum;
         branchtrans (q^.branchnum, 1-(1-q^.v)*(1-p^.next^.v));
         p^.next^.back := NIL;
         p^.next^.next^.back := NIL;
         update (q);
         update (q^.back);
      END; (* remove *)

      PROCEDURE copy ( VAR a, b : tree);
         (* copy a tree *)
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            (* copy a node *)
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO
            b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.trans := a.trans;
         b.transprod := a.transprod;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE buildnewtip (m : INTEGER; VAR tr : tree);
         (* set up a new tip and interior node it is connected to *)
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         p := tr.nodep[nextsp + numsp - 2];
         hookup (tr.nodep[m], p);
         p^.v := initialv;
         p^.back^.v := initialv;
         branchtrans (m, initialv);
         p^.branchnum := m;
         p^.next^.branchnum := p^.number;
         p^.next^.next^.branchnum := p^.number;
         p^.back^.branchnum := m;
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR tr : tree);
         (* set up and adjust branch lengths of a three-species tree *)
      BEGIN (* buildsimpletree *)
         hookup (tr.nodep[enterorder[1]], tr.nodep[enterorder[2]]);
         tr.nodep[enterorder[1]]^.v := initialv;
         tr.nodep[enterorder[2]]^.v := initialv;
         branchtrans (2, initialv);
         tr.nodep[1]^.branchnum := 2;
         tr.nodep[2]^.branchnum := 2;
         buildnewtip (enterorder[3], tr);
         insert(tr.nodep[enterorder[3]]^.back, tr.nodep[enterorder[2]]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
         (* try adding p at q, proceed recursively through tree *)
      BEGIN (* addtraverse *)
         tempp := curtree.trans;
         tempm := curtree.transprod;
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         remove(p, q);
         curtree.trans := tempp;
         curtree.transprod := tempm;
         IF (NOT q^.tip) AND contin
         THEN BEGIN
            addtraverse (p, q^.next^.back, contin);
            addtraverse (p, q^.next^.next^.back, contin);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p : nodeptr);
         (* rearranges the tree, globally or locally *)
         VAR q, r : nodeptr;
      BEGIN (* rearrange *)
         IF (NOT p^.tip) AND (NOT p^.back^.tip)
         THEN BEGIN
            r := p^.next^.next;
            remove (r, q);
            addtraverse (r, q^.next^.back, global AND (nextsp = numsp));
            addtraverse (r, q^.next^.next^.back, global AND (nextsp = numsp));
            copy (bestree, curtree);
            IF global AND (nextsp = numsp)
            THEN WRITE(OUTPUT, '.');
            IF global AND (nextsp = numsp) AND (NOT succeeded)
            THEN BEGIN
               IF r^.back^.tip
               THEN BEGIN
                  r := r^.next^.next;
                  remove (r, q);
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  copy (bestree, curtree);
                  END;
               END;
            END;
         IF NOT p^.tip
         THEN BEGIN
            rearrange (p^.next^.back);
            rearrange (p^.next^.next^.back);
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 300;
         VAR tipy, i : INTEGER;
            scale, tipmax : REAL;

         PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
            (* establishes coordinates of nodes *)
            VAR q, first, last: nodeptr;         
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := TRUNC(over*lengthsum+0.5);
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               IF lengthsum > tipmax
               THEN tipmax := lengthsum;
               END
            ELSE BEGIN
               q := p^.next; 
               REPEAT
                  coordinates (q^.back, lengthsum+q^.v);
                  q := q^.next;
               UNTIL ((p <> curtree.start^.back) AND (p = q))
                  OR ((p = curtree.start^.back) AND (p^.next = q));
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := TRUNC(over*lengthsum + 0.5);
               p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra : BOOLEAN;
               r, first, last : nodeptr;
               done : BOOLEAN;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            q := curtree.start^.back;
            IF (i = p^.ycoord) AND (p = curtree.start^.back)
            THEN WRITE(OUTPUT, '-')
            ELSE WRITE(OUTPUT, ' ');
            extra := FALSE;
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  r := p^.next;
                  done := FALSE;
                  REPEAT
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN
                        q := r^.back;
                        done := TRUE;
                        END;
                     r := r^.next;
                  UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
                     OR ((p = curtree.start^.back) AND (r = p^.next)));
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  IF p = curtree.start^.back THEN last := p^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
               IF (n < 2) AND (NOT q^.tip) THEN n := 2;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  WRITE(OUTPUT, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     IF q^.number-numsp >= 10
                     THEN WRITE(OUTPUT, q^.number-numsp:2)
                     ELSE WRITE(OUTPUT, '-', q^.number-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(OUTPUT, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (i = p^.ycoord) AND (p = curtree.start^.back)
                     THEN BEGIN
                        WRITE(OUTPUT, p^.number-numsp:1);
                        n := n - 1;
                        END;
                     IF (last^.ycoord > i) AND (first^.ycoord < i)
                        AND ((NOT (i = p^.ycoord)) OR (p = curtree.start))
                     THEN BEGIN
                        WRITE(OUTPUT, '!');
                        FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(OUTPUT, ' ');
                     END;
               IF q <> p THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         tipy := 1;
         tipmax := 0.0;
         coordinates (curtree.start^.back, 0.0);
         scale := 1.0/TRUNC(tipmax + 0.999);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;
         (* print out information on branches of tree *)

         PROCEDURE describe (p : nodeptr);
            (* print out information on one branch *)
            VAR i : INTEGER;
               q : nodeptr;
               s, sumlr : REAL;

            FUNCTION sigma(q : nodeptr) : REAL;
               (* get 1.95996 * approximate standard
                  error of branch length *)
               VAR sump, sumr, sums, sumc, p, pover3, pijk,
                     Qjk, liketerm, f, slopef, curvef : REAL;
                  i, j, k, m1, m2 : INTEGER;
                  binom1, binom2 : ARRAY[0..maxcutter] OF REAL;
                  Prob, slopeP, curveP : transmatrix;
                  r : nodeptr;
                  x1, x2 : sitelike;
            BEGIN (* sigma *)
               p := q^.v;
               pover3 := p / 3;
               FOR i := 0 TO sitelength DO BEGIN
                  binom1[0] := EXP((sitelength-i)*LN(1-p));
                  FOR k := 1 TO sitelength-i DO
                     binom1[k] := binom1[k-1] * (p/(1-p))
                                     * (sitelength-i-k+1)/k;
                  binom2[0] := EXP(i*LN(1-pover3));
                  FOR k := 1 TO i DO
                     binom2[k] := binom2[k-1] * (pover3/(1-pover3))*(i-k+1)/k;
                  FOR j := 0 TO sitelength DO BEGIN
                     sump := 0.0;
                     sums := 0.0;
                     sumc := 0.0;
                     IF (i-j) > 0
                     THEN m1 := i-j
                     ELSE m1 := 0;
                     IF (sitelength - j) < i
                     THEN m2 := sitelength - j
                     ELSE m2 := i;
                     FOR k := m1 TO m2 DO BEGIN
                        pijk := binom1[j-i+k] * binom2[k];
                        sump := sump + pijk;
                        sums := sums + pijk * ((j-i+2*k)/p
                                 - (sitelength-k)/(1-p) - (i-k)/(3-p));
                        sumc := sumc + pijk * (-(j-i+2*k)/SQR(p)
                           - (sitelength-k)/SQR(1-p) - (i-k)/SQR(3-p));
                        END;
                     Prob[i,j] := sump;
                     slopeP[i,j] := sums;
                     curveP[i,j] := sumc;
                     END;
                  END;
               sumlr := 0.0;
               sumc := 0.0;
               sums := 0.0;
               r := q^.back;
               FOR i := 1 TO endsite DO BEGIN
                  f := 0.0;
                  slopef := 0.0;
                  curvef := 0.0;
                  sumr := 0.0;
                  x1 := q^.x[i];
                  x2 := r^.x[i];
                  FOR j := 0 TO sitelength DO BEGIN
                     liketerm := pi[j] * x1[j];
                     sumr := sumr + liketerm * x2[j];
                     FOR k := 0 TO sitelength DO BEGIN
                        Qjk := liketerm * x2[k];
                        f := f + Qjk*Prob[j,k];
                        slopef := slopef + Qjk*slopeP[j,k];
                        curvef := curvef + Qjk*curveP[j,k];
                        END;
                     END;
                  sumlr := sumlr + weight[i] * LN (f / sumr);
                  sums := sums + weight[i] * slopef / f;
                  sumc := sumc + weight[i] * (curvef / f - SQR(slopef / f));
                  END;
               IF trunc8
               THEN BEGIN
                  f := 0.0;
                  slopef := 0.0;
                  curvef := 0.0;
                  sumr := 0.0;
                  x1 := q^.x[0];
                  x2 := r^.x[0];
                  FOR j := 0 TO sitelength DO BEGIN
                     liketerm := pi[j] * x1[j];
                     sumr := sumr + liketerm * x2[j];
                     FOR k := 0 TO sitelength DO BEGIN
                        Qjk := liketerm * x2[k];
                        f := f + Qjk*Prob[j,k];
                        slopef := slopef + Qjk*slopeP[j,k];
                        curvef := curvef + Qjk*curveP[j,k];
                        END;
                     END;
                  sumlr := sumlr + weightsum * LN ((1.0-sumr) / (1.0-f));
                  sums := sums + weightsum * slopef / (1.0-f);
                  sumc := sumc + weightsum * (curvef / (1.0-f)
                                 + SQR(slopef / (1.0-f)));
                  END;
               IF sumc > 1.0E-6
               THEN sigma := (-sums + SQR(sums-3.841*sumc)) / sumc
               ELSE sigma := -1.0;
            END; (* sigma *)

         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, (q^.number - numsp):4, '      ');
            WRITE(OUTPUT, '    ');
            IF p^.tip
            THEN FOR i := 1 TO nmlngth DO WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, (p^.number - numsp):4, '      ');
            IF q^.v >= 0.75
            THEN WRITE(OUTPUT, '     infinity')
            ELSE WRITE(OUTPUT, -0.75*LN(1-1.333333*q^.v):13:5);
            s := sigma(q);
            IF s < 0.0
            THEN WRITE(OUTPUT, '     (     zero,    infinity)')
            ELSE BEGIN
               WRITE(OUTPUT, '     (');
               IF q^.v - s <= 0.0
               THEN WRITE(OUTPUT, '     zero')
               ELSE WRITE(OUTPUT, -0.75*LN(1 - 1.333333*(q^.v + s)):9:5);
               WRITE(OUTPUT, ',');
               IF q^.v + s >= 0.75
               THEN WRITE(OUTPUT, '    infinity')
               ELSE WRITE(OUTPUT, -0.75*LN(1 - 1.333333*(q^.v - s)):12:5);
               WRITE(OUTPUT, ')')
               END;
            IF sumlr > 1.9205
            THEN WRITE(OUTPUT, ' *');
            IF sumlr > 2.995
            THEN WRITE(OUTPUT, '*');
            WRITELN(OUTPUT);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'remember: ');
         IF outgropt
         THEN WRITE(OUTPUT, '(although rooted by outgroup) ');
         WRITELN(OUTPUT, 'this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'Ln Likelihood = ', curtree.likelihood:11:5);
         WRITELN(OUTPUT);
         IF NOT usertree
         THEN WRITELN(OUTPUT, 'Examined ', numtrees:4, ' trees');
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, ' Between        And            Length');
         WRITELN(OUTPUT, '      Approx. Confidence Limits');
         WRITE(OUTPUT, ' -------        ---            ------');
         WRITELN(OUTPUT, '      ------- ---------- ------');
         describe (curtree.start^.back^.next^.back);
         describe (curtree.start^.back^.next^.next^.back);
         describe (curtree.start);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '     *  = significantly positive, P < 0.05');
         WRITELN(OUTPUT, '     ** = significantly positive, P < 0.01');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* summarize *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO nmlngth DO
               IF p^.nayme[i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            treeout (p^.next^.next^.back);
            IF p = curtree.start^.back
            THEN BEGIN
               WRITE(treefile, ',');
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            END;
         IF p^.v >= 1.0
         THEN x := -1.0
         ELSE x := - 0.75 * LN(1-1.333333*p^.v);
         IF x > 0.0
         THEN w := TRUNC(0.4343*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.4343*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = curtree.start^.back
         THEN WRITELN(treefile, ';')
         ELSE WRITE(treefile, ':', x:w+7:5);
      END; (* treeout *)

      PROCEDURE treeread;
         (* read a user-defined tree *)
         VAR ch : CHAR;
            p : nodeptr;
            i, j, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* read forward to next occurrence of character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            (* add one node to user-defined tree *)
            VAR q : nodeptr;
               i, j, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..nmlngth] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               n := nextnode;
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               FOR i := 0 TO endsite DO
                  FOR j := 0 TO sitelength DO BEGIN
                     q^.x[i][j] := 1.0;
                     q^.next^.x[i][j] := 1.0;
                     q^.next^.next^.x[i][j] := 1.0;
                     END;
               READ(infile, ch);
               END
            ELSE BEGIN
               FOR i := 1 TO nmlngth DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')')
                        OR (n > nmlngth));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO nmlngth DO
                     found := found AND (str[i] = curtree.nodep[n]^.nayme[i]);
                  IF found
                  THEN q := curtree.nodep[n]
                  ELSE n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE(OUTPUT, 'Cannot find species: ');
                  FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               hookup (p, q);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
            p^.branchnum := q^.number;
            q^.branchnum := q^.number;
            q^.v := initialv;
            p^.v := initialv;
            branchtrans (q^.branchnum, initialv);
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            FOR i := 0 TO endsite DO
               FOR j := 0 TO sitelength DO BEGIN
                  p^.x[i][j] := 1.0;
                  p^.next^.x[i][j] := 1.0;
                  p^.next^.next^.x[i][j] := 1.0;
                  END;
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         (* find maximum likelihood branch lengths of user tree *)
         VAR i : INTEGER;
            dummy : REAL;
      BEGIN (* treevaluate *)
         FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      setuppi;
      IF usertree
      THEN BEGIN
         READLN(infile, numtrees);
         WRITE(OUTPUT, 'User-defined tree');
         IF numtrees > 1
         THEN WRITE(OUTPUT, 's');
         WRITELN(OUTPUT, ':');
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            treevaluate;
            printree;
            summarize;
            END;
         WRITELN(OUTPUT);
         IF (numtrees > 1) AND (weightsum > 1)
         THEN BEGIN
            WRITE(OUTPUT, 'Tree    Ln L      Diff Ln L     Its S.D.');
            WRITELN(OUTPUT, '   Significantly worse?');
            WRITELN(OUTPUT);
            IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
            FOR i := 1 TO num DO BEGIN
               WRITE(OUTPUT, i:3, logl[i]:12:5);
               IF maxwhich = i
               THEN WRITELN(OUTPUT, '  <------ best')
               ELSE BEGIN
                  sum := 0.0;
                  sum2 := 0.0;
                  FOR j := 1 TO endsite DO BEGIN
                     sum := sum + weight[j] * (logf[maxwhich][j]-logf[i][j]);
                     sum2 := sum2 + weight[j]
                                 *SQR(logf[maxwhich][j]-logf[i][j]);
                     END;
                  sd := SQRT((weightsum/(weightsum-1))
                         *(sum2-SQR(sum/weightsum)));
                  WRITE(OUTPUT, (logl[i]-maxlogl):12:5, sd:12:4);
                  IF sum > 1.95996*sd
                  THEN WRITELN(OUTPUT, '           Yes')
                  ELSE WRITELN(OUTPUT, '           No');
                  END;
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            END;
         END
      ELSE BEGIN
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         WRITELN(OUTPUT, 'Adding species:');
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, curtree.nodep[enterorder[1]]^.nayme[i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, curtree.nodep[enterorder[2]]^.nayme[i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, curtree.nodep[enterorder[3]]^.nayme[i]);
         WRITELN(OUTPUT);
         nextsp := 3;
         buildsimpletree (curtree);
         curtree.start := curtree.nodep[enterorder[1]];
         numtrees := 1;
         nextsp := 4;
         WHILE nextsp <= numsp DO BEGIN
            buildnewtip (enterorder[nextsp], curtree);
            bestree.likelihood := -999999.0;
            addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                                   curtree.start^.back, TRUE);
            copy (bestree, curtree);
            WRITE(OUTPUT, '   ');
            FOR j := 1 TO nmlngth DO
               WRITE(OUTPUT,
                  curtree.nodep[enterorder[nextsp]]^.nayme[j]);
            WRITELN(OUTPUT);
            IF global AND (nextsp = numsp)
            THEN BEGIN
               WRITELN(OUTPUT, 'Doing global rearrangements');
               WRITE(OUTPUT, '   ');
               END;
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start^.back);
               END;
            IF (nextsp = numsp)
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               IF trout AND (nextsp = numsp)
               THEN BEGIN
                  treeout (curtree.start^.back);
                  WRITELN(OUTPUT, 'Tree also written onto file');
                  WRITELN(OUTPUT);
                  END;
               END;
            nextsp := nextsp + 1;
            END;
         END;
   END; (* maketree *)

BEGIN (* maximum likelihood phylogenies from restriction sites *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   getinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
      REWRITE(treefile);
      END;
   IF NOT anerror
   THEN maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);   Turbo Pascal and maybe others too *)
      END;
END. (* maximum likelihood phylogenies from restriction sites *)

